@@ -2,6 +2,7 @@ package Dancer::Plugin::RPC::JSONRPC;
 use v5.10;
 use Dancer ':syntax';
 use Dancer::Plugin;
+use Scalar::Util 'blessed';
 
 no if $] >= 5.018, warnings => 'experimental::smartmatch';
 
@@ -10,6 +11,7 @@ use Dancer::RPCPlugin::DispatchFromConfig;
 use Dancer::RPCPlugin::DispatchFromPod;
 use Dancer::RPCPlugin::DispatchItem;
 use Dancer::RPCPlugin::DispatchMethodList;
+use Dancer::RPCPlugin::FlattenData;
 
 my %dispatch_builder_map = (
     pod    => \&build_dispatcher_from_pod,
@@ -30,11 +32,13 @@ register jsonrpc => sub {
         }
     }
 
-    my $code_wrapper = $arguments->{code_wrapper} // sub {
-        my $code = shift;
-        my $pkg  = shift;
-        $code->(@_);
-    };
+    my $code_wrapper = $arguments->{code_wrapper}
+        ? $arguments->{code_wrapper}
+        : sub {
+            my $code = shift;
+            my $pkg  = shift;
+            $code->(@_);
+        };
     my $callback = $arguments->{callback};
     my $dispatcher = $publisher->($arguments->{arguments}, $endpoint);
 
@@ -58,10 +62,15 @@ register jsonrpc => sub {
             debug("[handle_jsonrpc_call] $method_name ", $request);
 
             if (!exists $dispatcher->{$method_name}) {
-                push @responses, jsonrpc_error_response(
-                    -32601,
-                    "Method '$method_name' not found",
-                    $request->{id}
+                push(
+                    @responses,
+                    jsonrpc_error_response(
+                        $request->{id},
+                        {
+                            code    => -32601,
+                            message => "Method '$method_name' not found",
+                        }
+                    )
                 );
                 next;
             }
@@ -74,18 +83,39 @@ register jsonrpc => sub {
             };
 
             if (my $error = $@) {
-                push @responses, jsonrpc_error_response(
-                    500,
-                    $error,
-                    $request->{id}
+                push(
+                    @responses,
+                    jsonrpc_error_response(
+                        $request->{id},
+                        {
+                            code    => 500,
+                            message => $error,
+                        }
+                    )
                 );
                 next;
             }
-            if (!$continue->success) {
+            if (!blessed($continue) || !$continue->isa('Dancer::RPCPlugin::CallbackResult')) {
                 push @responses, jsonrpc_error_response(
-                    $continue->error_code,
-                    $continue->error_message,
-                    $request->{id}
+                    $request->{id},
+                    {
+                        code    => -32603,
+                        message => "Internal error: 'callback_result' wrong class " . blessed($continue),
+                    }
+                );
+                next;
+            }
+
+            if (!$continue->success) {
+                push(
+                    @responses,
+                    jsonrpc_error_response(
+                        $request->{id},
+                        {
+                            code    => $continue->error_code,
+                            message => $continue->error_message,
+                        }
+                    )
                 );
                 next;
             }
@@ -97,23 +127,30 @@ register jsonrpc => sub {
             my $result = eval {
                 $code_wrapper->($handler, $package, $method_name, @method_args);
             };
+            my $error = $@;
 
             debug("[handeled_jsonrpc_call] ", $result);
-            if (my $error = $@) {
+            if ($error) {
                 push @responses, jsonrpc_error_response(
-                    500,
-                    $error,
-                    $request->{id}
+                    $request->{id},
+                    {
+                        code    => 500,
+                        message => $error,
+                    }
                 );
                 next;
             }
-            push @responses, {
-                jsonrpc => '2.0',
-                result => $result,
-                exists $request->{id}
-                    ? (id => $request->{id})
-                    : (),
-            };
+            if (blessed($result) && $result->can('as_jsonrpc_error')) {
+                push @responses, jsonrpc_error_response(
+                    $request->{id},
+                    $result->as_jsonrpc_error->{error}
+                );
+                next
+            }
+            elsif (blessed($result)) {
+                $result = flatten_data($result);
+            }
+            push @responses, jsonrpc_response($request->{id}, $result);
         }
 
         # create response
@@ -146,15 +183,22 @@ sub unjson {
     return @requests;
 }
 
-sub jsonrpc_error_response {
-    my ($code, $message, $id) = @_;
+sub jsonrpc_response {
+    my ($id, $data) = @_;
+
     return {
         jsonrpc => '2.0',
-        error => {
-            code    => $code,
-            message => $message,
-        },
-        defined $id ? (id => $id) : (),
+        id      => $id,
+        result  => $data,
+    };
+}
+
+sub jsonrpc_error_response {
+    my ($id, $error) = @_;
+    return {
+        jsonrpc => '2.0',
+        id      => $id,
+        error   => $error,
     };
 }
 
@@ -164,7 +208,7 @@ sub build_dispatcher_from_pod {
 
     return dispatch_table_from_pod(
         packages => $pkgs,
-        label    => 'xmlrpc',
+        label    => 'jsonrpc',
     );
 }
 
@@ -178,6 +222,7 @@ sub build_dispatcher_from_config {
         endpoint => $endpoint,
     );
 }
+
 register_plugin;
 1;
 
@@ -342,6 +387,10 @@ Deserializes the string as Perl-datastructure.
 
 Returns a jsonrpc error response as a hashref.
 
+=head2 jsonrpc_response
+
+Returns a jsonrpc response as a hashref.
+
 =head2 build_dispatcher_from_config
 
 Creates a (partial) dispatch table from data passed from the (YAML)-config file.
